#include <stdio.h>
#include<stdlib.h>
#include<math.h>
#include<sys/time.h>
#include<opencv/cv.h>
#include<opencv/highgui.h>
#include "graph.h"

#define max(a,b) ( ((a)>(b) ) ? (a) : (b) )
#define min(a,b) ( ((a)<(b) ) ? (a) : (b) )

//画像メモリ
IplImage *O_l,*O_r,*R_l,*R_r,*I_l,*I_r,*I_e,*F;
//画像サイズ
CvSize ori_size,size;
//階層ラベル
int A[256][256];
//リサイズ後の画像サイズ
int imagesize = 256;
//最大視差
int DISPMAX=64;
//SADのウィンドウサイズ
int WINSIZE=7;
//定数c_0,c_1
double c0=1.0;
double c1=1.0;
//閾値T
double T =10;

//現在時刻を取得する関数
double get_time_sec(){
  struct timeval tv;
  gettimeofday(&tv,NULL);
  return tv.tv_sec + (double)tv.tv_usec*1e-6;
}

//スムーズコストで使うcを計算
double C(int p_j,int p_i,int q_j,int q_i){
	//return c0;
	int edge = cvGet2D(I_e,p_j,p_i).val[0];
	if(edge > T)return c0;
	else return c1;//c0>>>c1
}

//データコストDを計算
double D(int j,int i,int d){

	if(i-d < 0 || size.width <= i-d)return 10000;

	int sad = 0;
	for(int s=-WINSIZE/2;s<WINSIZE/2;s++){
		if(j+s < 0 || size.height <= j+s)continue;
		for(int t=-WINSIZE/2;t<WINSIZE/2;t++){
			if(i+t < 0 || size.width <= i+t || 
				 i-d+t < 0 || size.width <= i-d+t)continue;
			sad += abs(cvGet2D(I_l,j+s,i+t).val[0]-cvGet2D(I_r,j+s,i-d+t).val[0]);
		}
	}
	
	return sad;
}

//スムーズコストVを計算
double V(int p_j,int p_i,int p,int q_j,int q_i,int q){
	return C(p_j,p_i,q_j,q_i) * abs(p-q);
}

//階層ラベルAを作る
void createA(int n){

	for(int i=0;i<256;i++){
		for(int j=0;j<256;j++){
			A[i][j] = -1;
		}
	}

	A[0][0] = 0;
	A[1][0] = n/2;
	int count = 2;
	for(int j=1;j<n;j++){
		double k_max = min(pow(2,j-1),n);
		//printf("%f\n",k_max);
	 
		for(int k=0;k<k_max;k++){
			int tmp = (1+4*k)*n/pow(2,j+1);
			if(tmp <= n){
				A[2*j][k] = (1+4*k)*n/pow(2,j+1);
				count++;
			}else break;
		}
		
		for(int k=0;k<k_max;k++){
			int tmp = (3+4*k)*n/pow(2,j+1);
			if(tmp <= n){
				A[2*j+1][k] = (3+4*k)*n/pow(2,j+1);
				count++;
			}else break;
		}
		if(count >= n)break;
	}

 	for(int i=0;i<n;i++){
		if(A[i][0] < 0)continue;
		printf("A[%d] ",i);
		for(int j=0;j<n;j++){
			if(A[i][j] < 0){
				printf("\n");
				break;
			}else printf("%d ",A[i][j]);
		}
	}
}

//エッジ検出を行う
void EdgeSearch(){

	double filter[3][3] = {
		{-2,-6,-2},
		{-6,32,-6},
		{-2,-6,-2}};

	for(int j=0;j<size.height;j++){
		for(int i=0;i<size.width;i++){
			int sum=0;
			for(int y=-1;y<2;y++){
				if(j+y < 0 || size.height <= j+y)continue;
				for(int x=-1;x<2;x++){
					if(i+x < 0 || size.width <= i+x )continue;
					sum += filter[y+1][x+1] * cvGet2D(I_r,j+y,i+x).val[0];
				}
			}
			//if(sum > 50)sum = 255;
			//else sum = 0; 
			//printf("%d\n",sum);
			cvSet2D(I_e,j,i,cvScalar(sum));
		}
	}
}

int main(int argc,char **argv){

	typedef Graph<int,int,int> GraphType;
	GraphType *g;

	printf("***************************************************************\n");
	
	//コマンドライン引数の数の確認
	if(argc != 8){
		printf("command line error >> Left_Image Right_image d c0 c1 T OutImage\n");
		exit(-1);
	}

	//入力画像の読み込み
	O_l = cvLoadImage(argv[1]);
	printf("left  image >> %s\n",argv[1]);
	O_r = cvLoadImage(argv[2]);
	printf("right image >> %s\n",argv[2]);
	ori_size = cvGetSize(O_r);
	//画像をリサイズ
	if(imagesize < ori_size.width){
		size.width = imagesize;
		size.height = ori_size.height * (double)imagesize/(double)ori_size.width+1;
	}else{
		size.width = ori_size.width;
		size.height = ori_size.height;
	}
	R_l = cvCreateImage(size,IPL_DEPTH_8U,3);
	cvResize(O_l,R_l,CV_INTER_NN);
	cvSaveImage("img/R_l.png",R_l,0);
	R_r = cvCreateImage(size,IPL_DEPTH_8U,3);
	cvResize(O_r,R_r,CV_INTER_NN);
	cvSaveImage("img/R_r.png",R_r,0);
	//グレースケール化
	I_l = cvCreateImage(size,IPL_DEPTH_8U,1);
	cvCvtColor(R_l,I_l,CV_RGB2GRAY);
	cvSaveImage("img/I_l.png",I_l,0);
	I_r = cvCreateImage(size,IPL_DEPTH_8U,1);
	cvCvtColor(R_r,I_r,CV_RGB2GRAY);
	cvSaveImage("img/I_r.png",I_r,0);
	//グラフのノード数とエッジ数を計算
	int node_pcs = size.width*size.height;
	int edge_pcs = (size.height-1)*size.width + (size.width-1)*size.height;	

	//パラメータの読み込み
	DISPMAX = atoi(argv[3]);
	createA(DISPMAX);
	printf("DISPMAX %d\n",DISPMAX);
	c0 = atof(argv[4]);
	printf("c0 %f\n",c0);
	c1 = atof(argv[5]);
	printf("c1 %f\n",c1);	
	T  = atof(argv[6]);
	printf("T  %f\n",T);
	
	//エッジ画像の作成
	I_e = cvCreateImage(size,IPL_DEPTH_8U,1);
	EdgeSearch();
	cvSaveImage("img/I_e.png",I_e,0);

	//総コストを初期化
	double Ep = 30000000;
	//各ノードpのAlpha_pを保持
	int Alpha[size.height][size.width],Beta[size.height][size.width];
	for(int j=0;j<size.height;j++){
		for(int i=0;i<size.width;i++){
			Beta[j][i] = 0;//DISPMAX-1;
		}
	}

	//計算開始時間の取得
	double stime = get_time_sec();

	//総コストEが収束するまでループ
	for(int loop=0;loop<100;loop++){

		int success = 0;
		double Et = Ep;
		int ave_node = 0,ave_edge = 0;

		//階層ラベルの深さだけループ
		for(int a=0;A[a][0]>=0;a++){

			fprintf(stderr,"loop %d >> %d / %d\r",loop,a,DISPMAX);
			
			//グラフの作成
			g = new GraphType(/*estimated # of nodes*/ 
												node_pcs 
												+ (size.width-1)*size.height
												+	size.width*(size.height-1),
												/*estimated # of edges*/ 
												edge_pcs + (size.width-1)*size.height + 
												size.width*(size.height-1)); 

			//ノードの設定
			for(int j=0;j<size.height;j++){
				for(int i=0;i<size.width;i++){
					g -> add_node(); 
					int beta_p  = Beta[j][i];
					int alpha_p = 0;
					for(int p=1;A[a][alpha_p+p]>=0;p++){
						if(abs(beta_p-A[a][alpha_p]) > abs(beta_p-A[a][alpha_p+p])){
							alpha_p += p;
							p = 1;
						}
					}
					Alpha[j][i] = alpha_p = A[a][alpha_p];

					g -> add_tweights
						( j*size.width+i, /* capacities */ D(j,i,alpha_p), D(j,i,beta_p) );
				}
			}
			
			//エッジの設定
			int n = node_pcs;
			for(int j=0;j<size.height;j++){
				for(int i=0;i<size.width;i++){

					int p = j*size.width+i;
					int beta_p  = Beta[j][i];
					int alpha_p = Alpha[j][i];

					//横のノードとのエッジを作る
					if(i+1 < size.width){
						int q = j*size.width+(i+1);
						int beta_q = Beta[j][i+1];
						int alpha_q = Alpha[j][i+1];
						int V_alpha_beta = V(j,i,alpha_p,j,i+1,beta_q);
						int V_beta_alpha = V(j,i,beta_p,j,i+1,alpha_q);
						
						if(alpha_p != alpha_q || beta_p != beta_q){
							//中間ノードを作る
							g -> add_node();
							int V_alpha = V(j,i,alpha_p, j,i+1,alpha_q);
							int V_beta  = V(j,i,beta_p , j,i+1,beta_q );
							g -> add_tweights( n, /* capacities */ V_alpha,V_beta );
							//エッジでつなぐ
							if(V_beta < V_alpha){
								g->add_edge( p, n, /* capacities */ 
														 max(0,(V_alpha_beta-V_beta)),10000 );
								g->add_edge( p, n, /* capacities */ 
														 max(0,(V_beta_alpha-V_beta)),10000 );
							}else{
								g->add_edge( p, n, /* capacities */ 
														 10000,max(0,(V_beta_alpha-V_alpha)) );
								g->add_edge( p, n, /* capacities */ 
														 10000,max(0,(V_alpha_beta-V_alpha)) );
							}
							n++;
						}else{
							//エッジでつなぐ
							g -> add_edge( p, q, V_alpha_beta, V_alpha_beta );
						}
					}
					
					//縦のノードとのエッジを作る
					if(j+1 < size.height){
						int q = (j+1)*size.width+i;
						int beta_q = Beta[j+1][i];
						int alpha_q = Alpha[j+1][i];
						int V_alpha_beta = V(j,i,alpha_p,j+1,i,beta_q);
						int V_beta_alpha = V(j,i,beta_p,j+1,i,alpha_q);

						if(alpha_p != alpha_q || beta_p != beta_q){
							//中間ノードを作る
							g -> add_node();
							int V_alpha = V(j,i,alpha_p, j+1,i,alpha_q);
							int V_beta  = V(j,i,beta_p , j+1,i,beta_q );
							g -> add_tweights( n, /* capacities */ V_alpha,V_beta );
							//エッジでつなぐ
							if(V_beta < V_alpha){
								g->add_edge( p, n, /* capacities */ 
														 max(0,(V_alpha_beta-V_beta)),10000 );
								g->add_edge( p, n, /* capacities */ 
														 max(0,(V_beta_alpha-V_beta)),10000 );
							}else{
								g->add_edge( p, n, /* capacities */ 
														 10000,max(0,(V_beta_alpha-V_alpha)) );
								g->add_edge( p, n, /* capacities */ 
														 10000,max(0,(V_alpha_beta-V_alpha)) );
							}
							n++;
						}else{
							//エッジでつなぐ
							g -> add_edge( p, q, V_alpha_beta, V_alpha_beta );
						}
					}

				}
			}
			ave_node += n;
			ave_edge += edge_pcs + (n-node_pcs);

			//グラフカット
			double E = g -> maxflow();
		
			if(E < Ep){
				//総コストの更新
				Ep = E;			
				//ラベルの更新
				for(int j=0;j<size.height;j++){
					for(int i=0;i<size.width;i++){
						if(g->what_segment(j*size.width+i) != GraphType::SOURCE){
							Beta[j][i] = Alpha[j][i];
						}
					}
				}
				success = 1;
			}

			//グラフの削除
			delete g;
		}

		printf("loop %d >> %d->%d %f %f\n",loop,(int)Et,(int)Ep,
					 ave_node/256.0,ave_edge/256.0);
		//総コストの値が変化しなかった場合はループを抜ける．
		if(success == 0)break;
	}

	double etime = get_time_sec();
	printf("run time %f[sec]\n",etime-stime);

	//出力画像の確保
	F = cvCreateImage(size,IPL_DEPTH_8U,1);
	for(int j=0;j<size.height;j++){
		for(int i=0;i<size.width;i++){
			cvSet2D(F,j,i,cvScalar(255*Beta[j][i]/DISPMAX));
		}
	}
	cvSaveImage(argv[7],F,0);
	printf("output image >> %s\n",argv[7]);

	//画像メモリの解放
	cvReleaseImage(&O_l);
	cvReleaseImage(&O_r);
	cvReleaseImage(&R_l);
	cvReleaseImage(&R_r);
  cvReleaseImage(&I_l);
	cvReleaseImage(&I_r);
	cvReleaseImage(&I_e);
  cvReleaseImage(&F);

	return 0;
}
